#!/usr/bin/env python
# coding: utf-8

from __future__ import annotations
from typing import Iterable

import re
import string
from collections import namedtuple

import numpy as np
import geopandas as gpd
import pandas as pd
import shapely.algorithms.polylabel as polylabel
import shapely.geometry as geom
import shapely.affinity as affine
import shapely.ops


"""Utility functions for use by various classes in the `weavingspace`
package. Most are geometric convenience functions for commonly applied
operations.
"""

PRECISION = 6
RESOLUTION = 1e-6

def _parse_strand_label(s:str) -> list[str]:
  """Breaks a strand label specification in to a list of labels

  Args:
    s (str): see get_strand_ids() for details
  Returns:
    list[str]: list of strand labels.
  """
  clean_s = re.sub("[(]+", "(", re.sub("[)]+", ")", s))
  is_combo = False
  output = []
  current = ""
  for c in clean_s:
    if is_combo:
      if c == ")":
        output.append(current)
        current = ""
        is_combo = False
      else:
        current = current + c
    else:
      if c == "(":
        is_combo = True
      else:
        output.append(c)
  return output


def get_strand_ids(strands_spec: str) -> tuple[list[str]]:
  """Converts a strands specification string to a list of lists of strand
  labels. String format is "a|bc|(de)f" where | separates strands in each
  direction and () designates combining labels into a single strand that
  will be sliced lengthwise. Example output:

        "a|bc|(de)f" -> (["a"], ["b", "c"], ["de", "f"])

  Superflous parentheses are removed, but no other error-checks are
  applied.

  Args:
    strands_spec (str): the strands specification to be parsed.

  Returns:
    tuple[str]: tuple of lists of labels for each set of strands.
  """
  strand_ids = [_parse_strand_label(s) for s in strands_spec.split("|")]
  strand_ids = strand_ids if len(strand_ids) == 3 else strand_ids + [[""]]
  return tuple(strand_ids)


def get_regular_polygon(spacing, n:int) -> geom.Polygon:
  r"""Returns regular polygon with n sides centered on (0, 0) with a horizontal
  base, and height given by spacing. Generation is based on equally spaced
  points around a circle with a radius calculated slightly differently 
  depend on whether the number of sides is even or odd.

            Even         Odd
        \  /    \  /      |
         \/   ___\/___    | 
         /\      /\      / \
        /  \    /  \    /   \

  The returned polygon is gridified. Note also that shapely silently reorders
  the coordinates clockwise from the lower left corner, regardless of the CCW
  order in which we supply them.

  Args:
    spacing (_type_): required height.
    n (int): number of sides.

  Returns:
    geom.Polygon: required geom.Polygon.
  """
  if n % 2 == 0:
    R = spacing / np.cos(np.pi / n) / 2
  else:
    R = spacing / (1 + np.cos(np.pi / n))
  angle_0 = -np.pi / 2 + np.pi / n
  a_diff = 2 * np.pi / n
  angles = [angle_0 + a_diff * i for i in range(n)]
  corners = [(R * np.cos(angle), R * np.sin(angle)) for angle in angles]
  return geom.Polygon(corners)


def get_corners(shape:geom.Polygon,
                repeat_first:bool = True) -> list[geom.Point]:
  """Returns a list of geom.Points around the boundary of a polygon, optionally
  repeating the first. Does no simplification (e.g. if a line segment has a 
  'corner' along its length, it is NOT removed; see get_clean_polygon for 
  that). Points have precision set to the package default tiling_utils.
  RESOLUTION.

  Args:
    shape (geom.Polygon): polygon whose corners are required.
    repeat_first (bool, optional): if True the first corner is repeated in the 
      returned list, if False it is omitted. Defaults to True.

  Returns:
    list[geom.Point]: list of geom.Point vertices of the polygon.
  """
  corners = [geom.Point(pt) for pt in shape.exterior.coords]
  return corners if repeat_first else corners[:-1]


def get_sides(shape:geom.Polygon) -> list[geom.LineString]:
  """Returns polygon sides as a list of geom.LineStrings, with resolution set
  to the package default tiling_utils.RESOLUTION. No simplification for 
  successive colinear sides is carried out.

  Args:
    shape (geom.Polygon): polygon whose edges are required.

  Returns:
    list[geom.LineString]: list of geom.LineString sides of the polygon.
  """
  corners = get_corners(shape)
  return [geom.LineString([p1, p2])
          for p1, p2 in zip(corners[:-1], corners[1:])]


def get_side_lengths(shape:geom.Polygon) -> list[float]:
  """Returns list of lengths of polygon sides. No simplification for corners
  along sides is carried out.

  Args:
    shape (geom.Polygon): polygon whose edge lengths are required.

  Returns:
    list[float]: list of side lengths.
  """
  corners = get_corners(shape)
  return [p1.distance(p2) for p1, p2 in zip(corners[:-1], corners[1:])]


def get_side_bearings(shape:geom.Polygon) -> tuple[tuple[float]]:
  """Returns a list of angles (in degrees) between the sides of a polygon and
  the positive x-axis, when proceeding from the first point in each side to its
  end point. This should usually be CW around the polygon.

  Args:
    shape (geom.Polygon): polygon whose side bearings are required.

  Returns:
    tuple[tuple[float]]: tuple of bearings of each edge.
  """
  return tuple([np.degrees(np.arctan2(
    e.coords[1][1] - e.coords[0][1], e.coords[1][0] - e.coords[0][0])) 
    for e in get_sides(shape)])


def get_interior_angles(shape:geom.Polygon) -> list[float]:
  """Returns angles (in degrees) between successive edges of a polygon. No 
  polygon simplification is carried out so some angles may be 180 (i.e. a 
  'corner' along a side, such that successive sides are colinear). 

  Args:
    shape (geom.Polygon): polygon whose angles are required.

  Returns:
    list[float]: list of angles.
  """
  corners = get_corners(shape, repeat_first = False)
  wrap_corners = corners[-1:] + corners + corners[:1]
  triples = zip(wrap_corners[:-2], wrap_corners[1:-1], wrap_corners[2:])
  return [get_inner_angle(p1, p2, p3) for p1, p2, p3 in triples]


def get_clean_polygon(
    shape:geom.MultiPolygon|geom.Polygon) -> geom.Polygon:
  """Returns polygon with any successive corners that are co-linear along a 
  side or very close to one another removed. 
  
  Particularly useful for tidying polygons weave tilings that have been 
  assembled from multiple 'cells' in the weave grid.

  Args:
    shape (geom.MultiPolygon|geom.Polygon): polygon to clean.

  Returns:
    geom.Polygon: cleaned polygon.
  """
  if isinstance(shape, geom.MultiPolygon):
    return geom.MultiPolygon([get_clean_polygon(p) for p in shape.geoms])
  elif isinstance(shape, geom.GeometryCollection):
    elements = [p for p in shape.geoms if isinstance(p, geom.Polygon)]
    if len(elements) > 1:
      return get_clean_polygon(geom.MultiPolygon(elements))
    else:
      return get_clean_polygon(elements[0])
  else:
    corners = get_corners(shape, repeat_first = False)
    # first remove any 'near neighbour' corners
    distances = get_side_lengths(shape)
    to_remove = [np.isclose(d, 0, rtol = RESOLUTION, atol = 10 * RESOLUTION) 
                 for d in distances]
    corners = [c for c, r in zip(corners, to_remove) if not r]
    # next remove any that are colinear
    p = geom.Polygon(corners)
    corners = get_corners(p, repeat_first = False)
    angles = get_interior_angles(p)
    to_remove = [np.isclose(a, 0, rtol = RESOLUTION, atol = RESOLUTION) or
                 np.isclose(a, 180, rtol = RESOLUTION, atol = RESOLUTION)
                 for a in angles]
    corners = [c for c, r in zip(corners, to_remove) if not r]
    return gridify(geom.Polygon(corners))


def rotate_preserving_order(
    polygon:geom.Polygon, 
    angle:float,
    centre:geom.Point = None) -> geom.Polygon:
  """
  Returns the supplied polygon rotated with the order of its corner points
    preserved (not guaranteed by shapely.affinity.rotate).

    Args:
        polygon (geom.Polygon): polygon to rotate.
        angle (float): desired angle of rotation (in degrees).
        centre (geom.Point): the rotation centre (passed on to shapely.affinity.
          rotate).

    Returns:
        geom.Polygon: rotated polygon.
  """
  ctr = "center" if centre is None else centre
  corners = get_corners(polygon, repeat_first = False)
  return geom.Polygon([affine.rotate(c, angle, ctr) for c in corners])


def ensure_cw(shape:geom.Polygon) -> geom.Polygon:
  """Returns the polygon with its outer boundary vertices in clockwise order.

  It is important to note that shapely.set_precision() imposes clockwise order
  on polygons, and since it is used widely throughout theses modules, it makes
  sense to impose this order.

    Args:
    shape (geom.Polygon): the polygon.

  Returns:
    geom.Polygon: the polygon in clockwise order.
  """
  if geom.LinearRing(shape.exterior.coords).is_ccw:
    return shape.reverse()
  else:
    return shape


def get_inner_angle(p1:geom.Point, p2:geom.Point, p3:geom.Point) -> float:
  r"""Returns the angle (in degrees) between line p1-p2 and p2-p3, i.e., the 
  angle A below
  
            p2
           / \ 
          / A \ 
        p1     p3
  
  Args:
    p1 (geom.Point): first point.
    p2 (geom.Point): second 'corner' point.
    p3 (geom.Point): third point.

  Returns:
      float: angle in degrees.
  """
  return np.degrees(np.arctan2(p3.y - p2.y, p3.x - p2.x) -
                    np.arctan2(p1.y - p2.y, p1.x - p2.x)) % 360


def is_regular_polygon(shape:geom.Polygon) -> bool:
  """Tests if supplied polygon is regular (i.e. equal sides and angles).

  Args:
      shape (geom.Polygon): polygon to test.

  Returns:
      bool: True if polygon is regular, False if not.
  """
  side_lengths = get_side_lengths(shape)
  angles = get_interior_angles(shape)
  return all(np.isclose(side_lengths, side_lengths[0])) \
     and all(np.isclose(angles, angles[0]))


def is_tangential(shape:geom.Polygon) -> bool:
  """Determines if the supplied polygon is tangential i.e., it can have circle
  inscribed tangential to all its sides. 

  NOTE: not currently used but retained in case incentre is restored to use
  'home made' code rather than polylabel.polylabel.
  
  Note that this will fail for polygons with successive colinear sides, meaning
  that polygons should be fully simplified...
  """
  if is_regular_polygon(shape):
    return True
  side_lengths = get_side_lengths(shape)
  n = len(side_lengths)
  if n % 2 == 1:
    if n == 3: #triangles are easy!
      return True
    # odd number of sides there is a nice solvable system  of equations see
    # https://math.stackexchange.com/questions/4065370/tangential-polygons-conditions-on-edge-lengths
    #
    # TODO: this is not reliable for reasons I don't fully understand. There is
    # a corresponding crude catch exception in incentre... but this needs a
    # more complete investigation
    mat = np.identity(n) + np.roll(np.identity(n), 1, axis = 1)
    fractions = np.linalg.inv(mat) @ side_lengths / side_lengths
    if not(np.isclose(np.mean(fractions), 
                      0.5, rtol= RESOLUTION, atol = RESOLUTION)):
      return False
    ones = [np.isclose(f, 1, rtol = RESOLUTION, atol = RESOLUTION) 
            for f in fractions]
    zeros = [np.isclose(f, 0, rtol = RESOLUTION, atol = RESOLUTION) 
             for f in fractions]
    negatives = [f < 0 for f in fractions]
    return not (any(ones) or any(zeros) or any(negatives))
  elif n == 4:
    # for quadrilaterals odd and even side lengths are equal
    return np.isclose(sum(side_lengths[0::2]), 
                      sum(side_lengths[1::2]), rtol = RESOLUTION)
  else:
    # other even numbers of sides... hard to avoid brute force
    bisectors = [get_angle_bisector(shape, i) for i in range(n)]
    intersects = [b1.intersection(b2) for b1, b2 in
                  zip(bisectors, bisectors[1:] + bisectors[:1])
                  if b1.intersects(b2)]
    distances = [i1.distance(i2) for i1, i2 in
                 zip(intersects, intersects[1:] + intersects[:1])]
    return all([d <= 2 * RESOLUTION for d in distances])


def incentre(shape:geom.Polygon) -> geom.Point:
  """A different polygon centre, which produces better results for some dual
  tilings where tiles are not regular polygons... see 
  https://en.wikipedia.org/wiki/Incenter

  This method relies on the polygon being tangential, i.e. there is an
  inscribed circle to which all sides of the polygon are tangent. It will work
  on all the polygons encountered in the Laves tilings, but is not guaranteed
  to work on all polygons.

  Given that the polygon is tangential, the radius of the inscribed circle is
  the [apothem of the polygon](https://en.wikipedia.org/wiki/Apothem) given by
  2 x Area / Perimeter. We apply a parallel offset of this size to two sides of
  the polygon and find their intersection to determine the centre of the circle
  givng the incentre of the polygon.

  Args:
    shape (geom.Polygon): the polygon.

  Returns:
    geom.Point: the incentre of the polygon.
  """
  if is_regular_polygon(shape):
    return shape.centroid
  return polylabel.polylabel(shape)
  # if is_regular_polygon(shape):
  #   return shape.centroid
  # shape = ensure_cw(shape)
  # if is_convex(shape):
  #   if is_tangential(shape):  # find the incentre
  #     corners = get_corners(shape, repeat_first = False)
  #     r = get_apothem_length(shape)
  #     e1 = geom.LineString(corners[:2]).parallel_offset(r, side = "right")
  #     e2 = geom.LineString(corners[1:3]).parallel_offset(r, side = "right")
  #     c = e1.intersection(e2)
  #     # is_tangential is unreliable, and sometimes we get to here and do not
  #     # get a Point, but a LineString, or even no intersection, so...
  #     return c if isinstance(c, geom.Point) else shape.centroid
  # return shape.centroid



def in_circle(shape:geom.Polygon) -> geom.Polygon:
  c = incentre(shape)
  r = min([c.distance(e) for e in get_sides(shape)])
  return c.buffer(r)


def get_apothem_length(shape:geom.Polygon) -> float:
  if is_regular_polygon(shape):
    return 2 * shape.area / geom.LineString(shape.exterior.coords).length
  else:
    c = polylabel.polylabel(shape, 1)
    return min([c.distance(e) for e in get_sides(shape)])


def get_angle_bisector(shape:geom.Polygon, v = 0) -> geom.LineString:
  """Returns a line which is the angle bisector of the specified corner of the
  supplied polygon.

  Args:
      shape (geom.Polygon): the polygon
      v (int, optional): index of the corner whose bisector is required. 
      Defaults to 0.

  Returns:
      geom.LineString: line which bisects the specified corner.
  """
  pts = get_corners(shape, repeat_first = False)
  n = len(pts)
  p0 = pts[v % n]
  p1 = pts[(v + 1) % n]
  p2 = pts[(v - 1) % n]
  d01 = p0.distance(p1)
  d02 = p0.distance(p2)
  if d01 < d02:
    p2 = geom.LineString([p0, p2]).interpolate(d01)
  else:
    p1 = geom.LineString([p0, p1]).interpolate(d02)
  m = geom.Point((p1.x + p2.x) / 2, (p1.y + p2.y) / 2)
  # we have to scale this to much larger in case of an angle near 180
  return affine.scale(geom.LineString([p0, m]), 10, 10, origin = p0)


def _get_interior_vertices(tiles:gpd.GeoDataFrame) -> gpd.GeoSeries:
  """Returns points not on the outer boundary of the supplied set
  of tiles.

  Args:
    shapes (gpd.GeoDataFrame): a set of polygons.

  Returns:
    gpd.GeoSeries: interior vertices of the set of polygons.
  """
  tiles = gridify(tiles.geometry)
  uu = safe_union(tiles, as_polygon = True)
  interior_pts = set()
  for tile in tiles:
    for corner in tile.exterior.coords:
      if uu.contains(geom.Point(corner).buffer(1e-3, cap_style = 3)):
        interior_pts.add(corner)
  return gridify(gpd.GeoSeries([geom.Point(p) for p in interior_pts]))


def gridify(
      gs:gpd.GeoSeries|gpd.GeoDataFrame|geom.Polygon
    ) -> gpd.GeoSeries|gpd.GeoDataFrame|geom.Polygon:
  """Returns the supplied GeoSeries rounded to the specified precision.

  Args:
    gs (gpd.GeoSeries): geometries to gridify.
    precision (int, optional): digits of precision. Defaults to 6.

  Returns:
    gpd.GeoSeries: the rounded geometries.
  """
  # return gpd.GeoSeries(
  #   list(gs.apply(
  #     wkt.dumps, rounding_precision = precision).apply(wkt.loads)))
  if isinstance(gs, (geom.Polygon, geom.Point, geom.MultiPoint,
                     geom.MultiPolygon, geom.LineString)) :
    return shapely.set_precision(gs, grid_size = RESOLUTION)
  if isinstance(gs, gpd.GeoSeries):
    return gpd.GeoSeries(
      [shapely.set_precision(g, grid_size = RESOLUTION) for g in list(gs)],
      crs = gs.crs)
  if isinstance(gs, gpd.GeoDataFrame):
    gs.geometry = gridify(gs.geometry)
    return gs


def get_dual_tile_unit(unit: TileUnit) -> gpd.GeoDataFrame:
  """Converts supplied TileUnit to a candidate GeoDataFrame of its dual
  TileUnit.

  NOTE: this is complicated and not remotely guaranteed to work!
  a particular issue is that where to place the vertices of the faces
  of the dual with respect to the tiles in the original is ill-defined.
  This is because the dual process is topologically not metrically defined,
  so that exact vertex locations are ambiguous. Tiling duality is defined in
  Section 4.2 of Grunbaum B, Shephard G C, 1987 _Tilings and Patterns_ (W. H.
  Freeman and Company, New York)

  NOTE: In general, this method will work only if all supplied tiles are
  regular polygons. A known exception is if the only non-regular polygons are
  triangles.

  NOTE: 'clean' polygons are required. If supplied polygons have messy
  vertices with multiple points where there is only one proper point, bad
  things are likely to happen! Consider using `clean_polygon()` on the
  tile geometries.

  Because of the above limitations, we only return a GeoDataFrame
  for inspection. However some `weavingspace.tile_unit.TileUnit` setup
  methods in `weavingspace.tiling_geometries` use this method, where we are
  confident the returned dual is valid.

  Args:
    unit (TileUnit): the tiling for which the dual is required.

  Returns:
    gpd.GeoDataFrame: GeoDataFrame that could be the tiles attribute for
      a TileUnit of the dual tiling.
  """
  # get a local patch of this Tiling
  local_patch = unit.get_local_patch(r = 1, include_0 = True)
  # Find the interior points of these tiles - these will be guaranteed
  # to have a sequence of surrounding tiles incident on them
  interior_pts = _get_interior_vertices(local_patch)
  # Compile a list of the polygons incident on the interior points
  cycles = []
  for pt in interior_pts:
    cycles.append(
      set([poly_id for poly_id, p in enumerate(local_patch.geometry)
           if pt.distance(p) < RESOLUTION * 2]))
  # convert the polygon ID sequences to (centroid.x, centroid.y, ID) tuples
  dual_faces = []
  for cycle in cycles:
    ids, pts = [], []
    for poly_id in cycle:
      ids.append(local_patch.tile_id[poly_id])
      poly = local_patch.geometry[poly_id]
      pts.append(incentre(poly))
    # sort them into CW order so they are well formed
    sorted_coords = _sort_cw([(pt.x, pt.y, id)
                              for pt, id in zip(pts, ids)])
    dual_faces.append(
      (geom.Polygon([(pt_id[0], pt_id[1]) for pt_id in sorted_coords]),
       "".join([pt_id[2] for pt_id in sorted_coords])))
  # ensure the resulting face centroids are inside the original tile
  # displaced a little to avoid uncertainties at corners/edges
  # TODO: Check  the logic of this - it seems like dumb luck that it works...
  dual_faces = [(f, id) for f, id in dual_faces
                if affine.translate(
                  unit.get_prototile_from_vectors().loc[0, "geometry"], 
                  RESOLUTION * 10, RESOLUTION * 10).contains(f.centroid)]
  gdf = gpd.GeoDataFrame(
    data = {"tile_id": [f[1] for f in dual_faces]}, crs = unit.crs,
    geometry = gpd.GeoSeries([f[0] for f in dual_faces]))
  # ensure no duplicates
  gdf = gdf.dissolve(by = "tile_id", as_index = False).explode(
    index_parts = False, ignore_index = True)
  gdf.tile_id = _relabel(gdf.tile_id)
  return gdf


def _relabel(data:Iterable) -> list:
  """Returns supplied data reassigned with unique values from
  string.ascii_letters.

  Args:
    data (Iterable): the data to relabel

  Returns:
    list: the reassigned data
  """
  new_data = {}
  d_count = 0
  for d in data:
    if not d in new_data:
      new_data[d] = string.ascii_letters[d_count]
      d_count = d_count + 1
  return [new_data[d] for d in data]


def _sort_cw(pts_ids:list[tuple[float, float, str]]):
  """Sorts supplied tuple of x, y, ID into clockwise order relative to their
  mean centre.

  Args:
    pts_ids (list[tuple[float, float, str]]): A tuple of a pair of
      floats and a string.

  Returns:
    list: a list in the same format as supplied sorted into
      clockwise order of the point locations.
  """
  x = [p[0] for p in pts_ids]
  y = [p[1] for p in pts_ids]
  cx, cy = np.mean(x), np.mean(y)
  dx = [_ - cx for _ in x]
  dy = [_ - cy for _ in y]
  angles = [np.arctan2(dy, dx) for dx, dy in zip(dx, dy)]
  d = dict(zip(angles, pts_ids))
  return [pt_id for angle, pt_id in reversed(sorted(d.items()))]


def write_map_to_layers(gdf:gpd.GeoDataFrame, fname:str = "output.gpkg",
                        tile_var:str = "tile_id") -> None:
  """Writes supplied GeoDataFrame to a GPKG file with layers based on
  the tile_var attribute.

  Args:
    gdf (gpd.GeoDataFrame): the GeoDataFrame.
    fname (str, optional): filename to write.
    tile_var (str, optional): the attribute to use to separate
      output file into layers. Defaults to "tile_id".
  """
  grouped = gdf.groupby(tile_var, as_index = False)
  for e in pd.Series.unique(gdf[tile_var]):
    grouped.get_group(e).to_file(fname, layer = e, driver = "GPKG")


def get_collapse_distance(geometry:geom.Polygon) -> float:
  """Returns the distance under which the supplied polygon will shrink
  to nothing if negatively buffered by that distance.

  Performs a binary search between an upper bound based on the radius of
  the circle of equal area to the polygon, and 0.

  Args:
    geometry (geom.Polygon): the polygon.

  Returns:
    float: its collapse distance.
  """
  if is_regular_polygon(geometry):
    return get_apothem_length(geometry)
  radius = np.sqrt(geometry.area / np.pi)
  lower = 0
  upper = radius
  delta = 1e12
  stop_delta = radius / 1000
  while delta > stop_delta:
    new_r = (lower + upper) / 2
    if geometry.buffer(-new_r).area > 0:
      lower = new_r
      delta = upper - new_r
    else:
      upper = new_r
      delta = new_r - lower
  return new_r


def get_largest_polygon(polygons:gpd.GeoSeries) -> gpd.GeoSeries:
  """Returns the largest polygon in a GeoSeries as a GeoSeries of one polygon.

  Args:
    polygons (gpd.GeoSeries): the set of polygons to pick from.

  Returns:
    gpd.GeoSeries: the largest polygon.
  """
  actual_polygons = [p for p in polygons.geometry
                     if isinstance(p, (geom.Polygon, geom.MultiPolygon))]
  areas = [p.area for p in actual_polygons]
  max_area = max(areas)
  return gpd.GeoSeries([actual_polygons[areas.index(max_area)]])


def touch_along_an_edge(p1:geom.Polygon, p2:geom.Polygon) -> bool:
  """Tests if two polygons touch along an edge.

  Checks that the intersection area of the two polygons buffered by
  a small amount is large enough to indicate that they neighbour at more
  than a corner.

  Args:
    p1 (geom.Polygon): First polygon
    p2 (geom.Polygon): Second polygon

  Returns:
    bool: True if they neighbour along an edge
  """
  return p1.buffer(RESOLUTION, join_style = 2, cap_style = 3) \
    .intersection(p2.buffer(RESOLUTION, join_style = 2, cap_style = 3)) \
      .area > 16 * RESOLUTION ** 2


def get_width_height_left_bottom(gs:gpd.GeoSeries) -> tuple[float]:
  """Returns width, height, left and bottom limits of a GeoSeries

  Args:
    gs (geopandas.GeoSeries): GeoSeries for which limits are required.

  Returns:
    tuple: four float values of width, height, left and bottom of gs.
  """
  bb = gs.total_bounds
  return (bb[2] - bb[0], bb[3] - bb[1], bb[0], bb[1])


def get_bounding_ellipse(
    shapes:gpd.GeoSeries,
    mag:float = 1.0) -> gpd.GeoSeries:
  """Returns an ellipse containing the supplied shapes.

  The method used is to calculate the size of square that would contain
  the shapes, if they had an aspect ratio 1, then stretch the circle in
  the x, y directions according to the actual aspect ratio of the shapes.

  Args:
    shapes (gpd.GeoSeries): the shapes to be contained.
    mag (float, optional): optionally increase the size of the returned
      ellipse by this scale factor. Defaults to 1.0.

  Returns:
    gpd.GeoSeries: the set of shapes.
  """
  w, h, l, b = get_width_height_left_bottom(shapes)
  c = geom.Point(l + w / 2, b + h / 2)
  r = min(w, h) * np.sqrt(2)
  circle = [c.buffer(r)]
  return gpd.GeoSeries(circle, crs = shapes.crs) \
    .scale(w / r * mag / np.sqrt(2), h / r * mag / np.sqrt(2), origin = c)


def get_tiling_edges(tiles:gpd.GeoSeries) -> gpd.GeoSeries:
  """Returns linestring GeoSeries from supplied polygon GeoSeries.

  This is used to allow display of edges of tiles in legend when they are
  masked by an ellipse (if we instead clip polygons then the ellipse edge
  will also show in the result.)

  Args:
    shapes (gpd.GeoSeries): Polygons to convert.

  Returns:
    gpd.GeoSeries: LineStrings from the supplied Polygons.
  """
  tiles = tiles.explode(ignore_index = True)
  edges = [geom.LineString(p.exterior.coords) for p in tiles]
  return gpd.GeoSeries(edges, crs = tiles.crs)


def get_polygon_sector(polygon:geom.Polygon, start:float = 0.0,
                       end:float = 1.0) -> geom.Polygon:
  """Returns a sector of the provided Polygon.

  The returned sector is a section of the polygon boundary between the
  normalized start and end positions, and including the polygon centroid.
  Should (probably) only be applied to convex polygons.

  Args:
    shape (geom.Polygon): the Polygon.
    start (float): normalized start position along the boundary. Defaults to
      0.
    end (float): normalized start position along the boundary. Defaults to
      1.

  Returns:
    geom.Polygon: the requested polygon sector.
  """
  if start == end:
    # must return a null polygon since the calling context
    # expects to get something back... which most likely
    # is needed to align with other data
    return geom.Polygon()
  if start * end < 0:
    # either side of 0/1 so assume required sector includes 0
    e1, e2 = min(start, end), max(start, end)
    arc1 = shapely.ops.substring(geom.LineString(polygon.exterior.coords),
                                 np.mod(e1, 1), 1, normalized = True)
    arc2 = shapely.ops.substring(geom.LineString(polygon.exterior.coords),
                                 0, e2, normalized = True)
    sector = geom.Polygon([polygon.centroid] +
               list(arc1.coords) +
               list(arc2.coords)[1:])
  else:
    arc = shapely.ops.substring(geom.LineString(polygon.exterior.coords),
                                start, end, normalized = True)
    sector = geom.Polygon([polygon.centroid] + list(arc.coords))
  return sector


def repair_polygon(
    polygon:geom.Polygon|geom.MultiPolygon|gpd.GeoSeries,
    shrink_then_grow:bool = True) -> geom.Polygon|gpd.GeoSeries:
  """Convenience function to 'repair' a shapely polyon or GeoSeries by applying
  a negative buffer then the same positive buffer.

  Optionally the buffer may be applied in the opposite order (i.e. grow then
  shrink). This operation may also convert a MultiPolygon that has some 'stray'
  parts to a Polygon.

  This is method is often unofficially recommended (on stackexchange etc.)
  even in the shapely docs, to resolve topology issues and extraneous
  additional vertices appearing when spatial operations are repeatedly
  applied.

  Args:
    p (geom.Polygon|gpd.GeoSeries): Polygon or GeoSeries to clean.
    res (float, optional): buffer size to use. Defaults to 1e-3.
    shrink_then_grow (bool, optional): if True the negative buffer is
      applied first, otherwise the buffer operations are applied in
      reverse. Defaults to True.

  Returns:
    geom.Polygon|gpd.GeoSeries: the cleaned Polygon or GeoSeries.
  """
  if type(polygon) is gpd.GeoSeries:
    return gpd.GeoSeries([repair_polygon(p, shrink_then_grow) 
                          for p in polygon]).set_crs(polygon.crs)
  else:
    if polygon is None:
      return None
    if shrink_then_grow:
      return polygon.buffer(
        -RESOLUTION * 10, join_style = 2, cap_style = 3).buffer(
        RESOLUTION * 10, join_style = 2, cap_style = 3)
    else:
      return polygon.buffer(
        RESOLUTION * 10, join_style = 2, cap_style = 3).buffer(
        -RESOLUTION * 10, join_style = 2, cap_style = 3)


def safe_union(gs:gpd.GeoSeries,
               as_polygon:bool = False) -> gpd.GeoSeries|geom.Polygon:
  """Unions the supplied GeoSeries of Polygons while buffering them to avoid
  gaps and odd internal floating edges. Optionally returns a Polygon or a
  GeoSeries.

  NOTE: APPEARS NOT TO BE VERY SAFE... find workarounds for the few occasions
  on which it is used.

  Frequently when unioning polygons that are ostensibly adjacent 'rogue'
  internal boundaries remain in the result. We can avoid this by buffering the
  polygons before unioning them, then reversing the buffer on the unioned
  shape.

  Args:
    gs (gpd.GeoSeries): the Polygons to union.
    res (float, optional): size of the buffer to use. Defaults to 1e-3.
    as_polygon (bool, optional): if True returns a Polygon, otherwise
      returns a one Polygon GeoSeries. Defaults to False.

  Returns:
    gpd.GeoSeries|geom.Polygon: the resulting union of supplied
      polygons.
  """
  r = RESOLUTION * 10
  union = gpd.GeoSeries([p.buffer(r, join_style = 2, cap_style = 3) for p in gs],
                         crs = gs.crs) \
      .union_all().buffer(-r, join_style = 2, cap_style = 3)
  if as_polygon:
    return union
  else:
    return gpd.GeoSeries([union], crs = gs.crs)


# def zigzag_between_points(
#     p0:geom.Point, p1: geom.Point, n:int, h:float = 1.0,  smoothness: int = 0):

#   template_steps = n * 2 + 1
#   r = p0.distance(p1)
  
#   x = np.linspace(0, n * np.pi, template_steps, endpoint = True)
#   y = [np.sin(x) for x in x]
#   s = interpolate.InterpolatedUnivariateSpline(x, y, k = 2)

#   spline_steps = (n + smoothness) * 2 + 1
#   xs = np.linspace(0, n * np.pi, spline_steps, endpoint = True)
#   ys = s(xs)
  
#   sfx = 1 / max(x) * r
#   sfy = h * r / 2
#   theta = np.arctan2(p1.y - p0.y, p1.x - p0.x)

#   ls = geom.LineString([geom.Point(x, y) for x, y in zip(xs, ys)])
#   ls = affine.translate(ls, 0, -(ls.bounds[1] + ls.bounds[3]) / 2)
#   ls = affine.scale(ls, xfact = sfx, yfact = sfy, origin = (0, 0))
#   ls = affine.rotate(ls, theta, (0, 0), use_radians = True)
#   x0, y0 = list(ls.coords)[0]
#   return affine.translate(ls, p0.x - x0, p0.y - y0)


def get_translation_transform(dx:float, dy:float) -> list[float]:
  """Returns the shapely affine transform tuple for a translation.

  Args:
      dx (float): translation distance in x direction.
      dy (float): translation distance in y direction.

  Returns:
    list[float]: a six item list of floats, per the shapely.affinity.
    affine_transform method, see 
      https://shapely.readthedocs.io/en/stable/manual.html#affine-transformations
  """
  return [1, 0, 0, 1, dx, dy]


def get_rotation_transform(
    angle:float, centre:tuple[float] = None) -> list[float]:
  """Returns the shapely affine transform tuple for a rotation, optionally
  about a supplied centre point.

  Args:
      angle (float): the angle of rotation (in degrees).
      centre (tuple[float], optional): An option centre location. Defaults to 
      None, which will in turn be converted to (0, 0).

  Returns:
    list[float]: a six item list of floats, per the shapely.affinity.
      affine_transform method, see 
        https://shapely.readthedocs.io/en/stable/manual.html#affine-transformations
  """
  if centre is None or np.allclose((0, 0), centre, atol = RESOLUTION):
    a = np.radians(angle)
    return [np.cos(a), -np.sin(a), np.sin(a), np.cos(a), 0, 0]
  
  dx, dy = centre
  t1 = get_translation_transform(-dx, -dy)
  r = get_rotation_transform(angle)
  t2 = get_translation_transform(dx, dy)
  return combine_transforms([t1, r, t2])


def get_reflection_transform(
    angle:float, centre:tuple[float] = None) -> list[float]:
  """Returns a shapely affine transform tuple that will reflect a shape
  in a line at the specified angle, optionally through a specified centre
  point.

  Args:
    angle (float): angle to the x-axis of the line of reflection.
    centre (tuple[float], optional): point through which the line of 
      reflection passes. Defaults to None, which
      will in turn be converted to (0, 0).

  Returns:
    list[float]: a six item list of floats, per the shapely.affinity.
      affine_transform method, see 
        https://shapely.readthedocs.io/en/stable/manual.html#affine-transformations
  """
  if centre is None or np.allclose((0, 0), centre, atol = RESOLUTION):
    A = 2 * np.radians(angle)
    return (np.cos(A), np.sin(A), np.sin(A), -np.cos(A), 0, 0)
  dx, dy = centre
  t1 = get_translation_transform(-dx, -dy)
  r = get_reflection_transform(angle)
  t2 = get_translation_transform(dx, dy)
  return combine_transforms([t1, r, t2])


def combine_transforms(transforms:list[list[float]]) -> list[float]:
  """Returns a shapely affine transform list that combines the listed
  sequence of transforms applied in order.

  Args:
    transforms (list[list[float]]): sequence of transforms to combine.

  Returns:
    list[float]: a transform tuple combining the supplied transforms applied
      in order, see 
        https://shapely.readthedocs.io/en/stable/manual.html#affine-transformations
  """
  result = np.identity(3)
  for t in transforms:
    result = as_numpy_matrix(t) @ result
  return as_shapely_transform(result)


def reverse_transform(transform:list[float]) -> list[float]:
  """Returns the inverse shapely affine transform of the supplied transform.

  Args:
    transform (list[float]): the transform for which the inverse is desired.

  Returns:
    list[float]: shapely affine transform tuple that will invert the supplied
      transform.
  """
  return as_shapely_transform(
    np.linalg.inv(as_numpy_matrix(transform)))


def as_shapely_transform(arr:np.array) -> list[float]:
  """Returns the shapely affine transform list equivalent to the supplied
  numpy matrix of a conventional augmented affine transform matrix.

  Args:
    arr (np.array): augmented affine transform matrix of the desired 
      transform.

  Returns:
    list[float]: desired shapely affine transform list of floats.
  """
  return [arr[0][0], arr[0][1], arr[1][0], arr[1][1], arr[0][2], arr[1][2]]


def as_numpy_matrix(transform:list[float]) -> np.array:
  """Converts the supplied shapely affine transform list to an augmented
  affine transform matrix in numpy array form. This makes combining transforms
  much easier.

  Args:
    transform (list[float]): the transform in shapely format.

  Returns:
    np.array: the transform in numpy matrix format.
  """
  return np.array([[transform[0], transform[1], transform[4]],
                   [transform[2], transform[3], transform[5]],
                   [           0,            0,            1]])


StraightLine = namedtuple("StraightLine", "A B C")
"""Named tuple to represent equation of a straight line in standard 
Ax + By + C = 0 form."""


def get_straight_line(p1:geom.Point, p2:geom.Point, 
                      perpendicular:bool = False) -> StraightLine:
  if perpendicular:
    ls = affine.rotate(geom.LineString([p1, p2]), 90)
    pts = [p for p in ls.coords]
    x1, y1 = pts[0]
    x2, y2 = pts[1]
  else:
    x1, y1 = p1.x, p1.y
    x2, y2 = p2.x, p2.y
  return StraightLine(y1 - y2, x2 - x1, x1 * y2 - x2 * y1)


def get_straight_line(p1:geom.Point, p2:geom.Point, 
                      perpendicular:bool = False) -> StraightLine:
  if perpendicular:
    ls = affine.rotate(geom.LineString([p1, p2]), 90)
    pts = [p for p in ls.coords]
    x1, y1 = pts[0]
    x2, y2 = pts[1]
  else:
    x1, y1 = p1.x, p1.y
    x2, y2 = p2.x, p2.y
  return StraightLine(y1 - y2, x2 - x1, x1 * y2 - x2 * y1)


def get_intersection(line1:StraightLine, 
                     line2:StraightLine) -> geom.Point:
  x_set, y_set = False, False
  denominator = line1.A * line2.B - line2.A * line1.B
  if np.isclose(line1.A, 0, atol = 1e-4, rtol = 1e-4):
    y = -line1.C / line1.B
    y_set = True
  elif np.isclose(line2.A, 0, atol = 1e-4, rtol = 1e-4):
    y = -line2.C / line2.B
    y_set = True
  if np.isclose(line1.B, 0, atol = 1e-4, rtol = 1e-4):
    x = -line1.C / line1.A
    x_set = True
  elif np.isclose(line2.B, 0, atol = 1e-4, rtol = 1e-4):
    x = -line2.C / line2.A
    x_set = True
  if np.isclose(denominator, 0, atol = 1e-4, rtol = 1e-4):
    return None
  x = x if x_set else (line1.B * line2.C - line2.B * line1.C) / denominator
  y = y if y_set else (line1.C * line2.A - line2.C * line1.A) / denominator
  return geom.Point(x, y)



#!/usr/bin/env python
# coding: utf-8

"""Implements `weavingspace.tileable.TileShape` and
`weavingspace.tileable.Tileable` the base classes for
`weavingspace.tile_unit.TileUnit` and `weavingspace.weave_unit.WeaveUnit`.

`Tileable` should not be called directly, but is instead accessed from the
`weavingspace.tile_unit.TileUnit` or `weavingspace.weave_unit.WeaveUnit`
constructor.

Several methods of `weavingspace.tileable.Tileable` are generally useful and
can be accessed through its subclasses.
"""

from enum import Enum
from dataclasses import dataclass
import copy

import matplotlib.pyplot as pyplot

import geopandas as gpd
import numpy as np
import shapely.geometry as geom
import shapely.affinity as affine

from weavingspace import tiling_utils

class TileShape(Enum):
  """The available base tile shapes.

  NOTE: the TRIANGLE type does not persist, but should be converted to a
  DIAMOND or HEXAGON type during `Tileable` construction.
  """
  RECTANGLE = "rectangle"
  HEXAGON = "hexagon"
  TRIANGLE = "triangle"
  DIAMOND = "diamond"


@dataclass
class Tileable:
  """Class to represent a tileable set of tile geometries.
  """

  tiles:gpd.GeoDataFrame = None
  """the geometries with associated `title_id` attribute encoding their 
  different colouring."""
  prototile:gpd.GeoDataFrame = None
  """the tileable polygon (rectangle, hexagon or diamond)"""
  spacing:float = 1000.0
  """the tile spacing effectively the resolution of the tiling. Defaults to
  1000"""
  base_shape:TileShape = TileShape.RECTANGLE
  """the tile shape. Defaults to 'RECTANGLE'"""
  vectors:dict[tuple[int], tuple[float]] = None
  """translation vector symmetries of the tiling"""
  regularised_prototile:gpd.GeoDataFrame = None
  """polygon containing the tiles of this tileable, usually a union of its
  tile polygons"""
  crs:int = 3857
  """coordinate reference system of the tile. Most often an ESPG code but
  any valid geopandas CRS specification is valid. Defaults to 3857 (i.e. Web
  Mercator)."""
  rotation:float = 0.0
  """cumulative rotation of the tileable."""
  debug:bool = False
  """if True prints debug messages. Defaults to False."""

  # Tileable constructor called by subclasses - should not be used directly
  def __init__(self, **kwargs):
    for k, v in kwargs.items():
      if isinstance(v, str):
        # make any string arguments lower case
        self.__dict__[k] = v.lower()
      else:
        self.__dict__[k] = v
    # delegate making the tiles back to the subclass _setup_tiles() method
    # which is implemented differently by TileUnit and WeavUnit. It will return
    # a message if there's a problem.
    # There might be a try... except... way to do this more 'properly', but we
    # prefer to return something even if it's not what was requested - along
    # with an explanation / suggestion
    message = self._setup_tiles()
    if message is not None: # there was a problem
      print(message)
      self._setup_default_tileable()
      return None
    else:
      self.prototile = self.get_prototile_from_vectors()
      self._setup_regularised_prototile()
      return None


  def setup_vectors(
      self, 
      *args:list[tuple[float]]) -> None:
    """Sets up translation vectors of a Tileable, from either two or three
    supplied tuples. Two non-parallel vectors are sufficient for a tiling to
    work, but usually three will be supplied for tiles with a hexagonal base
    tile. We also store the reverse vectors - this is for convenience when
    finding a 'local patch'. This method is preferred during Tileable
    initialisation.

    The vectors are stored in a dictionary indexed by their
    coordinates, e.g.

      {( 1,  0): ( 100, 0), ( 0,  1): (0,  100),
       (-1,  0): (-100, 0), ( 0, -1): (0, -100)}

    For a tileable of type `TileShape.HEXAGON`, the indexing tuples
    have three components. See https://www.redblobgames.com/grids/hexagons/
    """
    vectors = list(args)
    # extend list to include the inverse vectors too
    for v in args:
      vectors = vectors + [(-v[0], -v[1])]
    if len(vectors) == 4:
      i = [1, 0, -1,  0]
      j = [0, 1,  0, -1]
      self.vectors = {(i, j): v for i, j, v in zip(i, j, vectors)}
    else:
      i = [ 0,  1,  1,  0, -1, -1]
      j = [ 1,  0, -1, -1,  0,  1]
      k = [-1, -1,  0,  1,  1,  0]
      self.vectors = {(i, j, k): v for i, j, k, v in zip(i, j, k, vectors)}
    return None


  def get_vectors(
      self, 
      as_dict: bool = False
      ) -> dict[tuple[int],tuple[float]]|list[tuple[float]]:
    """
    Returns symmetry translation vectors as floating point pairs. Optionally 
    returns the vectors in a dictionary indexed by their coordinates, e.g.

      {( 1,  0): ( 100, 0), ( 0,  1): (0,  100),
       (-1,  0): (-100, 0), ( 0, -1): (0, -100)}

    Returns:
      dict[tuple[int],tuple[float]]|list[tuple[float]]: either the vectors as a
        list of float tuples, or a dictionary of those vectors indexed by 
        integer coordinate tuples.
    """
    if as_dict:
      return self.vectors
    else:
      return list(self.vectors.values())
    
  
  def get_prototile_from_vectors(self) -> gpd.GeoDataFrame:
    r"""Contructs and returns a prototile unit based on the current set vectors
    of the Tileable. For rectangular tilings the prototile is formed by points
    at diagonal corners defined by the halved vectors. By inspection, each edge
    of the prototile is the resultant of adding two of the four vectors.

      ----
     |\  /|
     | \/ |
     | /\ |
     |/  \|
      ----

    In the hexagonal case we form three such quadrilaterals (but don't halve the
    vectors, because we need the extended length) and intersect them to find a
    hexagonal shape. This guarantees that each vector will connect two opposite
    faces of the hexagon, as desired. This seems the most elegant approach by
    geometric construction.

    The prototile is not uniquely defined. The shape returned by this method is
    not guaranteed to be the most 'obvious' one that a human might construct!

    Returns:
      gpd.GeoDataFrame: A suitable prototile shape for the tiling wrapped in a 
        GeoDataFrame.
    """
    vecs = self.get_vectors()
    if len(vecs) == 4:
      v1, v2, v3, v4 = [(x / 2, y / 2) for (x, y) in vecs]
      prototile = geom.Polygon([(v1[0] + v2[0], v1[1] + v2[1]),
                                (v2[0] + v3[0], v2[1] + v3[1]),
                                (v3[0] + v4[0], v3[1] + v4[1]),
                                (v4[0] + v1[0], v4[1] + v1[1])])
    else:
      v1, v2, v3, v4, v5, v6 = vecs
      q1 = geom.Polygon([v1, v2, v4, v5]) 
      q2 = geom.Polygon([v2, v3, v5, v6]) 
      q3 = geom.Polygon([v3, v4, v6, v1])
      prototile = q3.intersection(q2).intersection(q1)
    return gpd.GeoDataFrame(
      geometry = gpd.GeoSeries([prototile]),
      crs = self.crs)
  

  def _regularise_tiles(self) -> None:
    """Combines separate tiles that share a tile_id value into
    single tiles, if they would end up touching after tiling.

    Also adjusts the `Tileable.regularised_prototile`
    attribute accordingly.
    """
    self.regularised_prototile = copy.deepcopy(self.prototile)
    # This preserves order while finding uniques, unlike list(set()).
    # Reordering ids might cause confusion when colour palettes
    # are not assigned explicitly to each id, but in the order
    # encountered in the tile_id Series of the GeoDataFrame.
    tiles, tile_ids = [], []
    ids = list(self.tiles.tile_id.unique())
    for id in ids:
      fragment_set = list(
        self.tiles[self.tiles.tile_id == id].geometry)
      merge_result = self._merge_fragments(fragment_set)
      tiles.extend(merge_result)
      tile_ids.extend([id] * len(merge_result))

    self.tiles = gpd.GeoDataFrame(
      data = {"tile_id": tile_ids},
      crs = self.crs,
      geometry = gpd.GeoSeries([tiling_utils.get_clean_polygon(t) 
                                for t in tiles]))

    self.regularised_prototile = \
      self.regularised_prototile.explode(ignore_index = True)
    if self.regularised_prototile.shape[0] > 1:
      self.regularised_prototile.geometry = tiling_utils.get_largest_polygon(
        self.regularised_prototile.geometry)
    return None


  def _merge_fragments(
      self, 
      fragments:list[geom.Polygon]) -> list[geom.Polygon]:
    """
    Merges a set of polygons based on testing if they touch when subjected
    to the translation vectors provided by `get_vectors()`.

    Called by `regularise_tiles()` to combine tiles in a tile unit that
    may be fragmented as supplied but will combine after tiling into single
    tiles. This step makes for more efficient implementation of the
    tiling of map regions, and also adds to the woven look in particular
    where it means that 'threads' go beyond the edges of the tile shapes.

    Args:
      fragments (list[geom.Polygon]): A set of polygons to merge.

    Returns:
      list[geom.Polygon]: A minimal list of merged polygons.
    """
    if len(fragments) == 1:
      return [f for f in fragments if not f.is_empty]
    fragments = [f for f in fragments if not f.is_empty]
    prototile = self.prototile.loc[0, "geometry"]
    reg_prototile = copy.deepcopy(
      self.regularised_prototile.loc[0, "geometry"])
    changes_made = True
    while changes_made:
      changes_made = False
      for v in self.vectors.values():
        # empty list to collect the new fragments
        # assembled in this iteration
        next_frags = []
        t_frags = [affine.translate(f, v[0], v[1]) for f in fragments]
        # build a set of any near matching pairs of
        # fragments and their translated copies
        matches = set()
        for i, f1 in enumerate(fragments):
          for j, f2, in enumerate(t_frags):
            if i < j and tiling_utils.touch_along_an_edge(f1, f2):
              matches.add((i, j))
        # determine which of these when unioned has the larger area in common
        # with the prototile
        done_frags = set()
        for i, j in matches:
          f1, f2 = fragments[i], t_frags[j]
          u1 = f1.buffer(tiling_utils.RESOLUTION * 2, 
                         join_style = 2, cap_style = 3).union(
            f2.buffer(tiling_utils.RESOLUTION * 2, 
                      join_style = 2, cap_style = 3))
          u2 = affine.translate(u1, -v[0], -v[1])
          if prototile.intersection(u1).area > prototile.intersection(u2).area:
            u1 = u1.buffer(-tiling_utils.RESOLUTION * 2, 
                           join_style = 2, cap_style = 3)
            u2 = u2.buffer(-tiling_utils.RESOLUTION * 2, 
                           join_style = 2, cap_style = 3)
            next_frags.append(u1)
            reg_prototile = reg_prototile.union(u1).difference(u2)
          else:
            u1 = u1.buffer(-tiling_utils.RESOLUTION * 2, 
                           join_style = 2, cap_style = 3)
            u2 = u2.buffer(-tiling_utils.RESOLUTION * 2, 
                           join_style = 2, cap_style = 3)
            next_frags.append(u2)
            reg_prototile = reg_prototile.union(u2).difference(u1)
          changes_made = True
          done_frags.add(i)
          done_frags.add(j)
        fragments = [f for i, f in enumerate(fragments)
                     if not (i in done_frags)] + next_frags
    self.regularised_prototile.loc[0, "geometry"] = reg_prototile
    # self.regularised_prototile.geometry[0] = reg_prototile
    return [f for f in fragments if not f.is_empty] # don't return any duds


  def get_local_patch(
      self, 
      r:int = 1,
      include_0:bool = False) -> gpd.GeoDataFrame:
    """Returns a GeoDataFrame with translated copies of the Tileable.

    The geodataframe takes the same form as the `Tileable.tile` attribute.

    Args:
      r (int, optional): the number of 'layers' out from the unit to
        which the translate copies will extendt. Defaults to `1`.
      include_0 (bool, optional): If True includes the Tileable itself at
        (0, 0). Defaults to `False`.

    Returns:
      gpd.GeoDataFrame: A GeoDataframe of the tiles extended by a number 
        of 'layers'.
    """
    # a dictionary of all the vectors we need, starting with (0, 0)
    three_vecs = len([k for k in self.vectors.keys()][0]) == 3
    vecs = (
      {(0, 0, 0): (0, 0)}
      if three_vecs
      # if self.base_shape in (TileShape.HEXAGON,)
      else {(0, 0): (0, 0)}
    )
    steps = r if three_vecs else r * 2
    # a dictionary of the last 'layer' of added vectors
    last_vecs = copy.deepcopy(vecs)
    # get the translation vectors in a dictionary indexed by coordinates
    # we keep track of the sum of vectors using the (integer) coordinates
    # to avoid duplication of moves due to floating point inaccuracies
    vectors = self.get_vectors(as_dict = True)
    for i in range(steps):
      new_vecs = {}
      for k1, v1 in last_vecs.items():
        for k2, v2 in vectors.items():
          # add the coordinates to make a new key...
          new_key = tuple([k1[i] + k2[i] for i in range(len(k1))])
          # if we haven't reached here before store the actual vector
          if not new_key in vecs:
            new_vecs[new_key] = (v1[0] + v2[0], v1[1] + v2[1])
      # extend the vectors and set the last layer to the set just added
      vecs = vecs | new_vecs
      last_vecs = new_vecs
    if not include_0:  # throw away the identity vector
      vecs.pop((0, 0, 0) if three_vecs else (0, 0))
    ids, tiles = [], []
    # we need to add the translated prototiles in order of their distance from 
    # tile 0, esp. in the square case, i.e. something like this:
    #
    #      5 4 3 4 5
    #      4 2 1 2 4
    #      3 1 0 1 3
    #      4 2 1 2 4
    #      5 4 3 4 5
    #
    # this is important for topology detection, where filtering back to the
    # local patch of radius 1 is simplified if prototiles have been added in 
    # this order. We use the vector index tuples not the euclidean distances
    # because this is more resistant to odd effects for non-convex tiles
    extent = self.get_prototile_from_vectors().loc[0, "geometry"]
    extent = affine.scale(extent, 
                          2 * r + tiling_utils.RESOLUTION, 
                          2 * r + tiling_utils.RESOLUTION)
    vector_lengths = {index: np.sqrt(sum([_ ** 2 for _ in index]))
                      for index in vecs.keys()}
    ordered_vector_keys = [k for k, v in sorted(vector_lengths.items(), 
                                                key = lambda item: item[1])]
    for k in ordered_vector_keys:
      v = vecs[k]
      if geom.Point(v[0], v[1]).within(extent):
        ids.extend(self.tiles.tile_id)
        tiles.extend(
          self.tiles.geometry.apply(affine.translate, xoff = v[0], yoff = v[1]))
    return gpd.GeoDataFrame(
      data = {"tile_id": ids}, crs=self.crs,
      geometry = gpd.GeoSeries(tiles))


  # applicable to both TileUnits and WeaveUnits
  def inset_tiles(
      self, 
      inset:float = 0) -> "Tileable":
    """Returns a new Tileable with an inset applied around the tiles.

    Works by applying a negative buffer of specfied size to all tiles.
    Tiles that collapse to zero area are removed and the tile_id
    attribute updated accordingly.

    NOTE: this method is likely to not preserve the relative area of tiles.

    Args:
      inset (float, optional): The distance to inset. Defaults to `0`.

    Returns:
      "Tileable": the new inset Tileable.
    """
    inset_tiles, inset_ids = [], []
    for p, id in zip(self.tiles.geometry, self.tiles.tile_id):
      b = p.buffer(-inset, join_style = 2, cap_style = 3)
      if not b.area <= 0:
        inset_tiles.append(b)
        inset_ids.append(id)
    result = copy.deepcopy(self)
    result.tiles = gpd.GeoDataFrame(
      data={"tile_id": inset_ids},
      crs=self.crs,
      geometry=gpd.GeoSeries(inset_tiles),
    )
    return result


  def scale_tiles(
      self, 
      sf:float = 1, 
      individually:bool = False) -> "Tileable":
    """Scales the tiles by the specified factor, centred on (0, 0).

    Args:
      sf (float, optional): scale factor to apply. Defaults to 1.

    Returns:
      TileUnit: the scaled TileUnit.
    """
    if individually:
      self.tiles.geometry = gpd.GeoSeries(
        [affine.scale(g, sf, sf) for g in self.tiles.geometry])
    else:
      self.tiles.geometry = self.tiles.geometry.scale(sf, sf, origin = (0, 0))
    return self


  def transform_scale(
      self, 
      xscale:float = 1.0, 
      yscale:float = 1.0,
      independent_of_tiling = False) -> "Tileable":
    """Transforms tileable by scaling.

    Args:
      xscale (float, optional): x scale factor. Defaults to 1.0.
      yscale (float, optional): y scale factor. Defaults to 1.0.

    Returns:
      Tileable: the transformed Tileable.
    """
    result = copy.deepcopy(self)
    result.tiles.geometry = self.tiles.geometry.scale(
      xscale, yscale, origin=(0, 0))
    if not independent_of_tiling:
      result.prototile.geometry = self.prototile.geometry.scale(
        xscale, yscale, origin=(0, 0))
      result.regularised_prototile.geometry = \
        self.regularised_prototile.geometry.scale(xscale, yscale, origin=(0, 0))
      result._set_vectors_from_prototile()
    return result


  def transform_rotate(
      self, 
      angle:float = 0.0,
      independent_of_tiling = False) -> "Tileable":
    """Transforms tiling by rotation.

    Args:
      angle (float, optional): angle to rotate by. Defaults to 0.0.

    Returns:
      Tileable: the transformed Tileable.
    """
    result = copy.deepcopy(self)
    result.tiles.geometry = self.tiles.geometry.rotate(angle, origin=(0, 0))
    if not independent_of_tiling:
      result.prototile.geometry = \
        self.prototile.geometry.rotate(angle, origin=(0, 0))
      result.regularised_prototile.geometry = \
        self.regularised_prototile.geometry.rotate(angle, origin=(0, 0))
      result._set_vectors_from_prototile()
      result.rotation = self.rotation + angle
    return result


  def transform_skew(
      self,
      xa:float = 0.0, 
      ya:float = 0.0,
      independent_of_tiling = False) -> "Tileable":
    """Transforms tiling by skewing

    Args:
      xa (float, optional): x direction skew. Defaults to 0.0.
      ya (float, optional): y direction skew. Defaults to 0.0.

    Returns:
      Tileable: the transformed Tileable.
    """
    result = copy.deepcopy(self)
    result.tiles.geometry = self.tiles.geometry.skew(xa, ya, origin=(0, 0))
    if not independent_of_tiling:
      result.prototile.geometry = \
        self.prototile.geometry.skew(xa, ya, origin=(0, 0))
      result.regularised_prototile.geometry = \
        self.regularised_prototile.geometry.skew(xa, ya, origin=(0, 0))
      result._set_vectors_from_prototile()
    return result


  def _set_vectors_from_prototile(self) -> None:
    """Sets the symmetry translation vectors as floating point pairs indexed
    by integer tuples with respect to either a rectangular or triangular grid 
    location. This method derives the vectors from a supplied prototile, and is
    intended to be used internally, after a transform_scale, _skew, or _rotate.

    These are 'face to face' vectors of the prototile, so that a hexagonal tile 
    will have 3 vectors, not the minimal parallelogram pair. Also sets the 
    inverse vectors. See `Tileable.setup_vectors()` for details.
    """
    t = self.prototile.loc[0, "geometry"]
    points = [p for p in t.exterior.coords][:-1] # each point once only no wrap
    n_pts = len(points)
    vec_dict = {}
    if n_pts == 4:
      vecs = [(q[0] - p[0], q[1] - p[1])
          for p, q in zip(points, points[1:] + points[:1])]
      i = [1, 0, -1,  0]
      j = [0, 1,  0, -1]
      vec_dict = {(i, j): v for i, j, v in zip(i, j, vecs)}
    elif n_pts == 6:
      vecs = [(q[0] - p[0], q[1] - p[1])
          for p, q in zip(points, points[2:] + points[:2])]
      # hex grid coordinates associated with each of the vectors
      i = [ 0,  1,  1,  0, -1, -1]
      j = [ 1,  0, -1, -1,  0,  1]
      k = [-1, -1,  0,  1,  1,  0]
      vec_dict = {(i, j, k): v for i, j, k, v in zip(i, j, k, vecs)}
    self.vectors = vec_dict
    return None


  def plot(
      self,
      ax:pyplot.Axes = None, 
      show_prototile:bool = True, 
      show_reg_prototile:bool = True, 
      show_ids:str = "tile_id",
      show_vectors:bool = False, 
      r:int = 0, 
      prototile_edgecolour:str = "k",
      reg_prototile_edgecolour:str = "r", 
      vector_edgecolour:str = "k",
      alpha:float = 1.0,
      r_alpha:float = 0.5,
      cmap:list[str] = None, 
      figsize:tuple[float] = (8, 8), 
      **kwargs) -> pyplot.axes:
    """Plots a representation of the Tileable on the supplied axis. **kwargs
    are passed on to matplotlib.plot()

    Args:
      ax (_type_, optional): matplotlib axis to draw to. Defaults to None.
      show_prototile (bool, optional): if `True` show the tile outline.
        Defaults to `True`.
      show_reg_prototile (bool, optional): if `True` show the regularised tile
        outline. Defaults to `True`.
      show_ids (str, optional): if `tile_id` show the tile_ids. If
        `id` show index number. If None or `''` don't label tiles.
        Defaults to `tile_id`.
      show_vectors (bool, optional): if `True` show the translation
        vectors (not the minimal pair, but those used by
        `get_local_patch()`). Defaults to `False`.
      r (int, optional): passed to `get_local_patch()` to show context if
        greater than 0. Defaults to `0`.
      r_alpha (float, optional): alpha setting for units other than the
        central one. Defaults to 0.5.
      prototile_edgecolour (str, optional): outline colour for the tile.
        Defaults to `"k"`.
      reg_prototile_edgecolour (str, optional): outline colour for the
        regularised. Defaults to `"r"`.
      vector_edgecolour (str, optional): colour for the translation vectors.
        Defaults to `"k"`.
      cmap (list[str], optional): colour map to apply to the central
        tiles. Defaults to `None`.
      figsize (tuple[float], optional): size of the figure.
        Defaults to `(8, 8)`.
    
    Returns:
      pyplot.axes: to which calling context may add things.
    """
    w = self.prototile.loc[0, "geometry"].bounds[2] - \
      self.prototile.loc[0, "geometry"].bounds[0]
    n_cols = len(set(self.tiles.tile_id))
    if cmap is None:
      cm = "Dark2" if n_cols <= 8 else "Paired"
    else:
      cm = cmap
    if ax is None:
      ax = self.tiles.plot(
        column="tile_id", cmap=cm, figsize=figsize, alpha = alpha, **kwargs)
    else:
      self.tiles.plot(
        ax=ax, column="tile_id", cmap=cm, figsize=figsize, **kwargs)
    if show_ids != None and show_ids != "":
      do_label = True
      if show_ids == "tile_id" or show_ids == True:
        labels = self.tiles.tile_id
      elif show_ids == "id":
        labels = [str(i) for i in range(self.tiles.shape[0])]
      else:
        do_label = False
      if do_label:
        for id, tile in zip(labels, self.tiles.geometry):
          ax.annotate(id, (tile.centroid.x, tile.centroid.y),
            ha = "center", va = "center", bbox = {"lw": 0, "fc": "#ffffff40"})
    if r > 0:
      self.get_local_patch(r=r).plot(
        ax = ax, column = "tile_id", alpha = r_alpha, cmap = cm, **kwargs)
    if show_prototile:
      self.prototile.plot(ax = ax, ec = prototile_edgecolour, lw = 0.5, 
                          fc = "#00000000", **kwargs)
    if show_vectors:  # note that arrows in mpl are dimensioned in plotspace
      vecs = self.get_vectors()
      for v in vecs[: len(vecs) // 2]:
        ax.arrow(0, 0, v[0], v[1], color = vector_edgecolour, width = w * 0.002,
          head_width = w * 0.05, length_includes_head = True, zorder = 3)
    if show_reg_prototile:
      self.regularised_prototile.plot(
        ax = ax, ec = reg_prototile_edgecolour, fc = "#00000000", 
        lw = 1.5, zorder = 2, **kwargs)
    return ax


  def _get_legend_tiles(self) -> gpd.GeoDataFrame:
    """Returns the tiles augmented by a rotation column.

    This base implementation may be overridden by specific tile unit types.
    In particular see `weavingspace.weave_unit.WeaveUnit._get_legend_tiles()`.

    Returns:
      gpd.GeoDataFrame: the tiles GeoDataFrame with a rotation column added.
    """
    tiles = copy.deepcopy(self.tiles)
    tiles["rotation"] = 0
    return tiles


  def _setup_default_tileable(self) -> None:
    """Sets up a default Tileable for when the TileUnit generators fail.
    """
    # if we've somehow got to hear without a base_shape, make it rectangular
    if self.base_shape is None:
      self.base_shape = TileShape.RECTANGLE
    if self.spacing is None:
      self.spacing = 1000
    match self.base_shape:
      case TileShape.HEXAGON:
        ids = ["a"]
        tiles = [tiling_utils.get_regular_polygon(self.spacing, 6)]
        self.setup_vectors(*self._get_hex_vectors())
      case TileShape.DIAMOND:
        ids = ["a"]
        tiles = [( self.spacing/2, 0), (0,  self.spacing * np.sqrt(3)/2),
                 (-self.spacing/2, 0), (0, -self.spacing * np.sqrt(3)/2)] 
        self.setup_vectors(*self._get_diamond_vectors())
      case TileShape.TRIANGLE:
        ids = ["a", "b"]
        t1 = tiling_utils.get_regular_polygon(self.spacing, 3)
        t1 = affine.translate(t1, 0, -t1.bounds[1])
        t2 = affine.rotate(t1, 180, origin = (0, 0))
        tiles = [t1, t2]
        self.setup_vectors(*self._get_diamond_vectors())
      case _:
        ids = ["a"]
        tiles = [tiling_utils.get_regular_polygon(self.spacing, 4)]
        self.setup_vectors(*self._get_square_vectors())
    self.tiles = gpd.GeoDataFrame(data = {"tile_id": ids},
                                  geometry = gpd.GeoSeries(tiles), 
                                  crs = 3857)
    self.crs = 3857
    self.prototile = self.get_prototile_from_vectors()
    self.regularised_prototile = copy.deepcopy(self.prototile)
    return None

  def _get_hex_vectors(self) -> list[tuple[float]]:
    """Helper function for `Tileable._setup_default_tileable()` returning three
    vectors for a hexagonal tiling.

    Returns:
        list[tuple[float]]: Translation vectors of a hexagonal tiling.
    """
    return [(v[0] * self.spacing, v[1] * self.spacing) 
            for v in [(0, 1), (np.sqrt(3)/2, 1/2), (np.sqrt(3)/2, -1/2)]]

  def _get_square_vectors(self):
    """Helper function for `Tileable._setup_default_tileable()` returning two
    vectors for a square tiling.

    Returns:
        list[tuple[float]]: Translation vectors of a square tiling.
    """
    return [(v[0] * self.spacing, v[1] * self.spacing) 
            for v in [(1, 0), (0, 1)]]

  def _get_diamond_vectors(self):
    """Helper function for `Tileable._setup_default_tileable()` returning two
    vectors for a diamond or triangular tiling.

    Returns:
        list[tuple[float]]: Translation vectors of a square tiling.
    """
    return [(v[0] * self.spacing, v[1] * self.spacing) 
            for v in [(1/np.sqrt(3), 1), (1/np.sqrt(3), -1)]]



#!/usr/bin/env python
# coding: utf-8

"""Functions to generate the matrices summarising tiles of tileable repeating 
geometries that when repeated across a map area give the appearance of a woven 
surface composed of criss-crossing strands. 

Implementation is based on ideas discussed in variously
 
Glassner A. 2002. Digital weaving. 1. IEEE Computer Graphics and Applications
22(6):108–118.
Glassner A. 2003. Digital weaving. 2. IEEE Computer Graphics and Applications 
23(1):77-90.
Glassner A. 2003. Digital weaving. 3. IEEE Computer Graphics and Applications 
23(2):80-83.
 
and (unpublished)

+ Griswold, R. 2006. Mathematical and Computational Topics in Weaving
https://www2.cs.arizona.edu/patterns/weaving/webdocs/mo/Griswold-MO.pdf 
(accessed 29/10/21).
 
where weaving is shown to be a matrix multiplication of tie-up, threading and 
treadling matrices. An accessible introduction can be found at 
https://www.youtube.com/watch?v=oMOSiag3dxg
"""

from __future__ import annotations
import numpy as np

def reps_needed(x1:int, x2:int) -> tuple[int]:
  """Returns how many repetitions of sequences of length x1 and x2 are 
  needed to make a matched length sequence 

  Args:
    x1 (int): length of first sequence.
    x2 (int): length of second sequence.

  Returns:
    tuple[int]: (number of repeats of x1, number of repeats of x2).
  """  
  n = np.lcm(x1, x2)
  return tuple(i for i in (n // x1, n // x2))


def get_pattern(tie_up:np.ndarray, treadling:np.ndarray, threading:np.ndarray,
        warp_n:int, weft_n:int, rep:int = 1) -> np.ndarray:
  """Returns a 0/1 encoded weave matrix. 
  
  Given tie_up, treadling and threading matrices. The following conditions 
  must be satisfied to avoid non-conformable matrix error:
  
    treadling.shape[1] == tie_up.shape[0]
    tie_up.shape[1] == threading.shape[0]
  
  The "twill", "basket" and "plain" options should guarantee this. If the 
  warp_n and weft_n values are not factors of treadling.shape[0] and 
  threading.shape[1] respectively, the output matrix will be repeated as needed
  to make this match

  Args:
    tie_up (np.ndarray): the tie up matrix.
    treadling (np.ndarray): the treadling matrix.
    threading (np.ndarray): the threading matrix.
    warp_n (int): the number of uniquely labelled warp threads.
    weft_n (int): the number of uniquely labelled weft threads.
    rep (int, optional): optional additional repetition of the output       
      applied in both directions. Defaults to 1.

  Returns:
    np.ndarray: the matrix weave pattern generated by the supplied parameters.
  """    
  pattern = (treadling @ tie_up @ threading > 0) + 0
  rep_warp = reps_needed(warp_n, pattern.shape[1])
  rep_weft = reps_needed(weft_n, pattern.shape[0])
  return np.tile(pattern, (rep_weft[1] * rep, rep_warp[1] * rep))


# Note that as currently written this function requires the warp and weft
# matrices to be the same size, which get_weave_pattern_matrix will ensure,
# but which may not be the case if called from elsewhere
def _encode_biaxial_weave(pattern:np.ndarray, warp:np.ndarray, 
             weft:np.ndarray) -> np.ndarray:
  """Encodes a biaxial weave pattern as below.

    1 - warp is absent
    2 - weft is absent
    3 - both threads are absent
    4 - weft is on top
    5 - warp is on top 
  
  Args:
    pattern (np.ndarray): pattern matrix with 1 where warp is on top, 
      0 where weft is on top.
    warp (np.ndarray): warp matrix where -1 values indicate absent threads.
    weft (np.ndarray): weft matrix where -1 values indicate absent threads.

  Returns:
    np.ndarray: matrix of values encoded as above.
  """    
  pattern = np.where(pattern == 1, 5, pattern)           # warp on top
  pattern = np.where(pattern == 0, 4, pattern)           # weft on top
  pattern = np.where(warp < 0, 1, pattern)               # warp absent
  pattern = np.where(weft < 0, 2, pattern)               # weft absent
  return np.where((warp < 0) & (weft < 0), 3, pattern)   # both absent


def make_plain_pattern(warp_n:int = 1, weft_n:int = 1) -> np.ndarray:
  """Returns plain weave (checkerboard) matrix.

  Args:
    warp_n (int, optional): number of warp thread labels. Defaults to 1.
    weft_n (int, optional): number of weft thread labels. Defaults to 1.

  Returns:
    np.ndarray: 0/1 matrix where 1 = warp on top.
  """    
  return make_twill_pattern(n = 1, warp_n = warp_n, weft_n = weft_n)


def make_twill_pattern(n:int|tuple[int] = 2, 
             warp_n:int = 2, weft_n:int = 2) -> np.ndarray:
  """Returns twill pattern matrix extended for warp and weft patterns.

  n is the number of over-unders. With n = 1 we get a plain weave.

  Args:
    n (int|tuple[int], optional): specifies over-under sequence in
      the weave. Defaults to 2.
    warp_n (int, optional): number of warp thread labels. Defaults to 2.
    weft_n (int, optional): number of weft thread labels. Defaults to 2.

  Returns:
    np.ndarray: _description_
  """    
  tie_up = make_twill_matrix(n)
  threading = np.diag(np.ones(tie_up.shape[0]))
  treadling = np.diag(np.ones(tie_up.shape[1]))
  return get_pattern(tie_up, treadling, threading, warp_n, weft_n)


def make_over_under_row(n:int|tuple[int]) -> list[int]:
  """Returns a tuple of runs of 1s and 0s.

  Returns a tuple of runs of 1s and 0s per supplied n. 
  
  If n is an integer, returned tuple will be a series of n 1s followed by
  n 0s.
  
  If n is a tuple of odd length it will be repeated to produce an even
  length over-under sequence. This is required to avoid cases such as
  e.g, (1,2,3) -> 100111 which repeated is 1001111001111, i.e. a (2,4)
  pattern. Repeating it yields 100111011000 which is the requested pattern.

  Args:
    n (int|tuple[int]): requested over-under sequence. See details.

  Returns:
    list[int]: list of runs of 1s and 0s in the requested pattern. 
  """    
  over_under = n
  if type(over_under) == int:
    over_under = [n, n]
  elif len(n) % 2 != 0:
    over_under = n * 2
  x = 1
  row = []
  for y in over_under:
    row.extend([x] * y)
    x = 1 - x
  return row


def wrap_row(r:list, by:int = 1) -> list:
  """Wraps the list r by number of positions.
  
  Positive by will shift right.
  Negative shifts left.

  Args:
    r (list): list to wrap.
    by (int, optional): number of positions to shift by. Defaults to 1.

  Returns:
    list: wrapped list.
  """    
  return r[-by:] + r[:-by]


def make_twill_matrix(over_under:int|tuple[int]) -> np.ndarray:
  """Makes a twill 0/1 matrix.

  Makes a matrix like
  
    1 1 0 0
    0 1 1 0
    0 0 1 1
    1 0 0 1
    
  where the repeat runs in each row are lengths returned by
  make_over_under_rown(n)

  Args:
    over_under (int|tuple[int]): over-under run specification. See
      make_over_under_row().

  Returns:
    np.ndarray: a matrix of 0s and 1s.
  """    
  row = make_over_under_row(over_under)
  d = len(row)
  out = []
  for i in range(d):
    row = wrap_row(row, 1)
    out.extend(row)
  return np.array(out).reshape(d, d)


def make_basket_pattern(n:int = 2, warp_n:int = 2, 
            weft_n:int = 2) -> np.ndarray: 
  """Returns basket pattern matrix extended for warp and weft patterns.

  Args:
    n (int, optional): over under count. Defaults to 2.
    warp_n (int, optional): number of warp thread labels. Defaults to 2.
    weft_n (int, optional): number of weft thread labels. Defaults to 2.

  Returns:
    np.ndarray: _description_
  """    
  tie_up = make_basket_matrix(n)
  threading = np.diag(np.ones(tie_up.shape[0]))
  treadling = np.diag(np.ones(tie_up.shape[1]))
  return get_pattern(tie_up, treadling, threading, warp_n, weft_n)


def make_basket_matrix(n:int) -> np.ndarray:
  """Returns a basket weave pattern matrix.

  Makes a matrix like
  
    1 1 0 0
    1 1 0 0
    0 0 1 1
    0 0 1 1
  
  where the repeat runs in each row are length n

  Args:
    n (int): dimension of the 'basket' over-under pattern.

  Returns:
    np.ndarray: 0/1 matrix in basket weave pattern.
  """    
  return np.array((([1] * n + [0] * n) * n) + 
          (([0] * n + [1] * n) * n)).reshape(n * 2, n * 2)


def make_this_pattern(tie_up:np.ndarray, 
            threading:np.ndarray = None,
            treadling:np.ndarray = None,
            warp_n:int = 1, weft_n:int = 1) -> np.ndarray:
  """Pass through returns weave pattern matrix from supplied input.

  This is just a pass through function which applies suitable size identity. 
  Could try to enforce
  
    treadling.shape[1] == tie_up.shape[0] and 
    tie_up.shape[1] == threading.shape[0]
  
  but unsure what would be an appropriate way to do this...

  Args:
    tie_up (np.ndarray): desired weave pattern.
    threading (np.ndarray): desired threading pattern.
    treadling (np.ndarray): desired treadling pattern.
    warp_n (int, optional): number of warp thread labels. Defaults to 1.
    weft_n (int, optional): number of weft thread labels. Defaults to 1.

  Returns:
    np.ndarray: resulting 0/1 weave pattern matrix.
  """    
  threading = (np.diag(np.ones(tie_up.shape[0]))
         if threading is None
         else threading)
  treadling = (np.diag(np.ones(tie_up.shape[1]))
         if treadling is None
         else treadling)
  return get_pattern(tie_up, treadling, threading, warp_n, weft_n)


def get_weave_pattern_matrix(weave_type:str = "plain", 
               n:int|tuple[int] = 2, 
               warp:list[str]|tuple[str] = ["a", "b"],
               weft:list[str]|tuple[str] = ["c", "d"], 
               tie_up:np.ndarray = make_twill_matrix((2, 2)), 
               tr:np.ndarray = None, 
               th:np.ndarray = None) -> np.ndarray:
  """Returns encoded weave pattern matrix.
  
  See `_encode_biaxial_weave()` for the encoding.

  Allowed weave_types: "plain", "twill", "basket", and "this" (pass-thru).
  These are explained in the respective functions to which this function
  delegates construction of the base matrices before applying the encoding.

  Args:
    weave_type (str, optional): one of "plain", "twill", "basket" or
      "this". Defaults to "plain".
    n (int|tuple[int], optional): over under pattern. See 
      make_over_under_row() for details. Defaults to 2.
    warp (list[str]|tuple[str], optional): list of labels for warp 
      strands. Defaults to ["a", "b"].
    weft (list[str]|tuple[str], optional): list of labels for weft 
      strands. Defaults to ["c", "d"].
    tie_up (np.ndarray, optional): a weave pattern matrix to pass thru in 
      the "this" case. Defaults to make_twill_matrix((2, 2)).
    tr (np.ndarray, optional): treadling matrix for the "this" case.        
      Defaults to None.
    th (np.ndarray, optional): threading matrix for the "this" case.        
      Defaults to None.

  Returns:
    np.ndarray: encoded weave pattern matrix.
  """    
  warps = [-1 if c == "-" else i for i, c in enumerate(warp)]
  wefts = [-1 if c == "-" else i for i, c in enumerate(weft)]
  width = len(warp)
  height = len(weft)
  
  if weave_type == "plain":
    p = make_plain_pattern(warp_n = width, weft_n = height)
  elif weave_type == "twill":
    p = make_twill_pattern(n = n, warp_n = width, weft_n = height)
  elif weave_type == "basket":
    p = make_basket_pattern(n = n, warp_n = width, weft_n = height)
  else:
    p = make_this_pattern(tie_up = tie_up, threading = th, treadling = tr,
                warp_n = width, weft_n = height)
  nr, nc = p.shape
  warp_threads = np.array(warps * 
      reps_needed(nc, len(warps))[1] * nr).reshape((nr, nc))
  weft_threads = np.array(wefts * 
      reps_needed(nr, len(wefts))[1] * nc).reshape((nc, nr)).transpose()
  # encode to reflect missing threads
  return _encode_biaxial_weave(p, warp_threads, weft_threads)




#!/usr/bin/env python
# coding: utf-8

"""The `WeaveUnit` subclass of `weavingspace.tileable.Tileable` implements
tileable geometric patterns constructed by specifying 2- or 3-axial weaves.

Examples:
  Explain usage here...
"""

from __future__ import annotations
import itertools
import logging
from dataclasses import dataclass
from typing import Iterable

import pandas as pd
import geopandas as gpd
import numpy as np
import shapely.geometry as geom
import shapely.affinity as affine
import shapely.ops

from weavingspace import weave_matrices
from weavingspace import tiling_utils

from weavingspace import Loom
from weavingspace import WeaveGrid

from weavingspace import TileShape
from weavingspace import Tileable


@dataclass
class WeaveUnit(Tileable):
  """Extends Tileable to allow for tiles that appear like woven patterns.
  """

  weave_type:str = "plain"
  """type of weave pattern, one of `plain`, `twill`, `basket`, `cube`, `hex` or
  `this`. Defaults to `plain`."""
  aspect:float = 1.
  """width of strands relative to the `spacing`. Defaults to 1.0."""
  n:int|tuple[int] = (2, 2)
  """number of over-under strands in biaxial weaves. Only one item is 
  required in a plain weave. Twill and basket patterns expect an even number of 
  entries in the tuple."""
  strands:str = "a|b|c"
  """specification of the strand labels along each axis. Defaults to `a|b|c`."""
  _tie_up:np.ndarray = None
  """optional tie-up array to pass through for `this` weave type."""
  _tr:np.ndarray = None
  """optional treadling array to pass through for `this` weave type."""
  _th:np.ndarray = None
  """optional threading array to pass through for `this` weave type."""

  def __init__(self, **kwargs):
    super(WeaveUnit, self).__init__(**kwargs)


  def _setup_tiles(self) -> None:
    """Returns dictionary with weave unit and tile GeoDataFrames based on
    parameters already supplied to the constructor.
    """
    self._parameter_info()

    if self.weave_type in ("hex", "cube"):
      self.base_shape = TileShape.HEXAGON
      self._setup_triaxial_weave_unit()
    else:
      self.base_shape = TileShape.RECTANGLE
      self._setup_biaxial_weave_unit()
    return


  def _setup_regularised_prototile(self) -> None:
    """Sets up a 'regularised prototile' which fully contains all the tile
    elements, i.e. it does not intersect. The work is carried out by the two
    methods  that follow _regularise_tiles() and _merge_fragments().
    """
    self._regularise_tiles()
    # it's prudent to do some cleanup given all the manipulation of geometries
    # carried out to generate the regularised prototile. But note that the 
    # regularised prototile has no functional purpose, so it's OK if it has,
    # for example, additional points along line segments.
    self.regularised_prototile.geometry = tiling_utils.repair_polygon(
      self.regularised_prototile.geometry)
    return None


  def _parameter_info(self) -> None:
    """Outputs logging message concerning the supplied aspect settings.
    """
    if self.aspect == 0:
      logging.info("Setting aspect to 0 is probably not a great plan.")

    if self.aspect < 0 or self.aspect > 1:
      logging.warning(
        """Values of aspect outside the range 0 to 1 won't produce tiles 
        that will look like weaves, but they might be pretty anyway! Values
        less than -1 seem particularly promising, especially with opacity 
        set less than 1.""")

    return None


  def _setup_biaxial_weave_unit(self) -> None:
    """Returns weave tiles GeoDataFrame and tile GeoDataFrame in a 
    dictionary based on paramters already supplied to constructor.
    """
    warp_threads, weft_threads, _ = \
      tiling_utils.get_strand_ids(self.strands)
    if self.weave_type == "basket" and isinstance(self.n, (list, tuple)):
      self.n = self.n[0]
    p = weave_matrices.get_weave_pattern_matrix(
      weave_type = self.weave_type, n = self.n, warp = warp_threads,
      weft = weft_threads, tie_up = self._tie_up, tr = self._tr,
      th = self._th)
    self._make_shapes_from_coded_weave_matrix(
      Loom(p), strand_labels = [weft_threads, warp_threads, []])
    bb = self.tiles.total_bounds
    w = (bb[2] - bb[0]) // self.spacing * self.spacing
    h = (bb[3] - bb[1]) // self.spacing * self.spacing
    self.setup_vectors((0, h), (w, 0))
    return None


  def _get_triaxial_weave_matrices(self,
    strands_1:list[str]|tuple[str] = ["a"],
    strands_2:list[str]|tuple[str] = ["b"],
    strands_3:list[str]|tuple[str] = ["c"]) -> Loom:
    """Returns encoded weave pattern matrix as Loom of three biaxial matrices.

    Allowed weave_types: "cube" or "hex".

    "hex" is not flexible and will fail with any strand label lists that are 
    not length 3 or include more than one non-blank "-" item. You can 
    generate the "hex" weave with the default settings in any case!

    Strand lists should be length 3 or length 1. "cube" tolerates more 
    than "hex" for the items in the strand lists.

    Defaults will produce 'mad weave'.

    Args:
      strands_1 (list[str]|tuple[str], optional): list of labels
      for warp strands. Defaults to ["a"].
      strands_2 (list[str]|tuple[str], optional): list of labels
      for weft strands. Defaults to ["b"].
      strands_3 (list[str]|tuple[str], optional): list of labels
      for weft strands. Defaults to ["c"].

    Returns:
      Loom: which combines the three biaxial weaves 12, 23 and 31 implied
      by the strand label lists.
    """
    if self.weave_type == "hex":
      loom = Loom(
        weave_matrices.get_weave_pattern_matrix(
          weave_type = "this", tie_up = np.ones((6, 6)),
          warp = strands_1, weft = strands_2),
        weave_matrices.get_weave_pattern_matrix(
          weave_type = "this", tie_up = np.ones((6, 6)),
          warp = strands_2, weft = strands_3),
        weave_matrices.get_weave_pattern_matrix(
          weave_type = "this", tie_up = np.ones((6, 6)),
          warp = strands_3, weft = strands_1))
    else: # "cube"
      loom = Loom(
      # Note n = (1,2,1,2) is required here to force 6x6 twill
        weave_matrices.get_weave_pattern_matrix(
          weave_type = "twill", n = (1, 2, 1, 2),
          warp = strands_1, weft = strands_2),
        weave_matrices.get_weave_pattern_matrix(
          weave_type = "twill", n = (1, 2, 1, 2),
          warp = strands_2, weft = strands_3),
        weave_matrices.get_weave_pattern_matrix(
          weave_type = "twill", n = (1, 2, 1, 2),
          warp = strands_3, weft = strands_1))
    return loom


  def _setup_triaxial_weave_unit(self) -> None:
    """Returns weave tiles GeoDataFrame and tile GeoDataFrame in a
    dictionary based on parameters already supplied to constructor.
    """
    strands_1, strands_2, strands_3 = \
      tiling_utils.get_strand_ids(self.strands)
    loom = self._get_triaxial_weave_matrices(
      strands_1 = strands_1, strands_2 = strands_2, strands_3 = strands_3)
    self._make_shapes_from_coded_weave_matrix(
      loom, strand_labels = [strands_1, strands_2, strands_3])
    self.setup_vectors((0, 6 * self.spacing),
                       (3 * self.spacing * np.sqrt(3),  3 * self.spacing),
                       (3 * self.spacing * np.sqrt(3), -3 * self.spacing))
    return None


  def _make_shapes_from_coded_weave_matrix(
    self, loom:Loom, strand_labels:list[list[str]] = [["a"], ["b"], ["c"]]
    ) -> None:
    """Returns weave tiles and prototile GeoDataFrames in a dictionary

    Builds the geometries associated with a given weave supplied as
    'loom' containing the coordinates in an appropriate grid (Cartesian or
    triangular) and the orderings of the strands at each coordinate location

    Args:
      loom (Loom): matrix or stack of matrices representing the weave
      pattern.
      strand_labels (list[list[str]], optional): list of lists of labels
      for strands in each direction. Defaults to [["a"], ["b"], ["c"]]
    """
    grid = WeaveGrid(loom.n_axes, loom.orientations, self.spacing)
    # expand the list of strand labels if needed in each direction
    # labels = []
    labels = [thread * int(np.ceil(dim // len(thread)))
              for dim, thread in zip(loom.dimensions, strand_labels)]
    weave_polys = [] 
    strand_ids = [] 
    cells = []
    for k, strand_order in zip(loom.indices, loom.orderings):
      ids = [thread[coord] for coord, thread in zip(k, labels)]
      cells.append(grid.get_grid_cell_at(k))
      if strand_order is None:
        continue  # No strands present
      if strand_order == "NA":
        continue  # Inconsistency in layer order
      n_slices = [len(id) for id in ids]
      next_polys = grid.get_visible_cell_strands(
        width = self.aspect, coords = k,
        strand_order = strand_order, n_slices = n_slices)
      weave_polys.extend(next_polys)
      next_labels = [list(ids[i]) for i in strand_order]  # list of lists
      next_labels = list(itertools.chain(*next_labels))   # flatten
      strand_ids.extend(next_labels)
    # sometimes empty polygons make it to here, so
    # filter those out along with the associated IDs
    real_polys = [not p.is_empty for p in weave_polys]
    weave_polys = [p for p, b in zip(weave_polys, real_polys) if b]
    strand_ids = [id for id, b in zip(strand_ids, real_polys) if b]
    # note that the tile is important for the biaxial case, which makes it
    # a little hard to understand why the behaviour is so different in 
    # the biaxial and triaxial cases; however below seems to work...
    if loom.n_axes == 3:
      tile = grid.get_tile_from_cells(cells)
      shift = (0, 0)
    else:
      tile = tiling_utils.safe_union(gpd.GeoSeries(cells), as_polygon = True)
      shift = (-tile.centroid.x, -tile.centroid.y)
      tile = grid.get_tile_from_cells(tile)
    self.tiles = self._get_weave_tiles_gdf(weave_polys, strand_ids, shift)
    self.prototile = gpd.GeoDataFrame(
      geometry = gpd.GeoSeries([tile]), crs = self.crs)
    # self.setup_vectors()
    return None


  def _get_weave_tiles_gdf(
      self, polys:list[geom.Polygon], strand_ids:list[str], 
      offset:tuple[float]) -> gpd.GeoDataFrame:
    """Makes a GeoDataFrame from weave tile polygons, labels, etc.

    Args:
      polys (list[Polygon | MultiPolygon]): list of weave tile
        polygons.
      strand_ids (list[str]): list of strand labels.
      offset (tuple[float]): offset to centre the weave tiles on the
        tile.

    Returns:
      geopandas.GeoDataFrame: GeoDataFrame clipped to the tile, with
        margin applied.
    """
    weave = gpd.GeoDataFrame(
      data = {"tile_id": strand_ids},
      geometry = gpd.GeoSeries([affine.translate(p, offset[0], offset[1])
                                for p in polys]))
    weave = weave[weave.tile_id != "-"]
    weave.geometry = gpd.GeoSeries(
      [tiling_utils.get_clean_polygon(p) for p in weave.geometry])
    
    # some buffering is required if aspect is 1 to safely dissolve and
    # explode weave unit tiles that meet at corners
    if self.aspect == 1:
      # grow for dissolve
      weave.geometry = weave.geometry.buffer(
        # self.spacing * tiling_utils.RESOLUTION, 
        tiling_utils.RESOLUTION, 
        join_style = 2, cap_style = 3)
      weave = weave.dissolve(by = "tile_id", as_index = False)
      # shrink by more to explode into separate polygons
      weave.geometry = weave.geometry.buffer(
        # -2 * self.spacing * tiling_utils.RESOLUTION, 
        -2 * tiling_utils.RESOLUTION, 
        join_style = 2, cap_style = 3)
      weave = weave.explode(ignore_index = True)
      weave.geometry = weave.geometry.buffer(
        # self.spacing * tiling_utils.RESOLUTION, 
        tiling_utils.RESOLUTION, 
        join_style = 2, cap_style = 3)
    else: # aspect < 1 is fine without buffering
      weave = weave.dissolve(by = "tile_id", as_index = False)
      weave = weave.explode(ignore_index = True)
    weave.geometry = gpd.GeoSeries(
      [tiling_utils.get_clean_polygon(p) for p in weave.geometry])
    return weave.set_crs(self.crs)


  def _get_axis_from_label(self, label:str = "a", strands:str = None):
    """Determines the axis of a tile_id from the strands spec string.

    Args:
      label (str, optional): the tile_id. Defaults to "a".
      strands (str, optional): the strand spec. Defaults to the WeaveUnit
      strands attribute.

    Returns:
      _type_: the axis in which the supplied tile is found.
    """
    if strands == None:
      strands = self.strands
    index = strands.index(label)
    return strands[:index].count("|")


  def _get_legend_tiles(self) -> gpd.GeoDataFrame:
    """Returns tiles suitable for use in a legend representation.

    One tile for each tile_id value will be chosen, close to the
    centre of the prototile extent, and not among the smallest tiles present
    (for example not a short length of strand mostly hidden by other
    strands)

    Returns:
      gpd.GeoDataFrame: the chosen tiles.
    """
    angles = ((0, 240, 120)
              if self.weave_type in ("hex", "cube")
              else (90, 0))
    tile_ids = pd.Series.unique(self.tiles.tile_id)
    groups = self.tiles.groupby("tile_id")
    tiles, rotations = [], []
    for id in tile_ids:
      candidates = groups.get_group(id)
      axis = self._get_axis_from_label(id, self.strands)
      tiles.append(
        self._get_most_central_large_tile(candidates, tiles))
      rotations.append(-angles[axis] + self.rotation)
    return gpd.GeoDataFrame(
      data = {"tile_id": tile_ids, "rotation": rotations},
      crs = self.crs,
      geometry = gpd.GeoSeries(tiles))


  def _get_most_central_large_tile(self, tiles:gpd.GeoDataFrame,
          other_tiles:list[geom.Polygon]) -> geom.Polygon:
    """Gets a large tile close to the centre of the WeaveUnit.

    Args:
      tiles (gpd.GeoDataFrame): the set of tiles to choose from.

    Returns:
      geom.Polygon: the chosen, large central tile.
    """
    areas = [g.area for g in tiles.geometry]
    min_area, max_area = min(areas), max(areas)
    if min_area / max_area > 0.5:
      geoms = list(tiles.geometry)
    else:
      mean_log_a = np.mean(np.log(areas))
      geoms = [g for g, a in zip(tiles.geometry, areas)
              if np.log(a) > mean_log_a]
    if len(other_tiles) == 0 or self.weave_type in ("cube", "hex"):
      d = [g.centroid.distance(geom.Point(0, 0)) for g in geoms]
    else:
      c = geom.MultiPolygon(other_tiles).centroid
      d = [geom.MultiPolygon([g] + other_tiles).centroid.distance(c)
          for g in geoms]
    return geoms[d.index(min(d))]


  def _get_legend_key_shapes(self, 
                             polygon:geom.Polygon,
                             counts:Iterable = [1] * 25, 
                             angle:float = 0,
                             radial:bool = False) -> list[geom.Polygon]:
    """Returns a list of polygons obtained by slicing the supplied polygon
    across its length into n slices. Orientation of the polygon is
    indicated by the angle.

    The returned list of polygons can be used to form a colour ramp in a
    legend.

    Args:
      polygon (geom.Polygon): the weave strand polygon to slice.
      counts (Iterable, optional): an iterable list of the numbers of
        slices in each category. Defaults to [1] * 25.
      angle (float, optional): orientation of the polygon. Defaults to 0.
      categorical (bool, optional): ignored by WeaveUnit.

    Returns:
      list[geom.Polygon]: a list of polygons.
    """
    c = polygon.centroid
    g = affine.rotate(polygon, -angle, origin = c)
    width, height, left, bottom = \
      tiling_utils.get_width_height_left_bottom(gpd.GeoSeries([g]))
    margin = width / 50
    total = sum(counts)
    cuts = list(np.cumsum(counts))
    cuts = [0] + [c / total for c in cuts]
    cuts = [left + c * width for c in cuts]
    bottom = bottom - margin
    top = bottom + height + 2 * margin
    slices = []
    for l, r in zip(cuts[:-1], cuts[1:]):
      # we add a margin to left and right so that they overplot; otherwise in
      # rendering matplotlib leaves small gaps which give a washed out look
      # to the fill colour!
      slice = geom.Polygon([(l - margin, bottom), (r + margin, bottom), 
                            (r + margin,    top), (l - margin,    top)])
      slices.append(slice.intersection(g))
    return [affine.rotate(s, angle, origin = c) for s in slices]


